﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

using DevExpress.XtraBars.Docking;
using DevExpress.XtraBars.Docking2010;
using DevExpress.XtraBars.Docking2010.Views;
using DevExpress.XtraBars.Docking2010.Views.Tabbed;

namespace TestProject
{
    /// <summary>
    /// 문서 컨트롤
    /// </summary>
    public partial class DocumentControl : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region 사각형 무효화 하기 - InvalidateRect(windowHandle, rectangleHandle, erase)

        /// <summary>
        /// 사각형 무효화 하기
        /// </summary>
        /// <param name="windowHandle">윈도우 핸들</param>
        /// <param name="rectangleHandle">사각형 핸들</param>
        /// <param name="erase">지우기 여부</param>
        /// <returns>처리 결과</returns>
        [DllImport("user32.dll")]
        private static extern bool InvalidateRect(IntPtr windowHandle, IntPtr rectangleHandle, bool erase);

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 도킹 패널 딕셔너리
        /// </summary>
        private Dictionary<string, DockPanel> dictionary = new Dictionary<string, DockPanel>();

        /// <summary>
        /// 문서 추가 여부
        /// </summary>
        private bool isAddDocument = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - DocumentControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public DocumentControl()
        {
            InitializeComponent();

            this.tabbedView.DocumentAdded += tabbedView_DocumentAdded;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 문서 추가하기 - AddDocument(name, text, contentControl, documentOption, dockStyle)

        /// <summary>
        /// 문서 추가하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <param name="text">텍스트</param>
        /// <param name="contentControl">컨텐트 컨트롤</param>
        /// <param name="documentOption">문서 옵션</param>
        /// <param name="dockStyle">도킹 스타일</param>
        /// <returns>처리 결과</returns>
        public bool AddDocument
        (
            string         name,
            string         text,
            Control        contentControl,
            DocumentOption documentOption = null,
            DockStyle      dockStyle = DockStyle.Fill
        )
        {
            try
            {
                this.tabbedView.BeginUpdate();
                this.dockManager.BeginUpdate();

                this.isAddDocument = true;

                if(this.dictionary.ContainsKey(name))
                {
                    return false;
                }

                DockingStyle dockingStyle = GetDockingStyle(dockStyle);

                DockPanel dockPanel = this.dockManager.AddPanel(DockingStyle.Left);

                dockPanel.Name                 = name;
                dockPanel.Text                 = text;
                dockPanel.Appearance.BackColor = Color.White;

                if(documentOption != null)
                {
                    SetDockPanel(dockPanel, documentOption);
                }

                CustomHeaderButton button = GetCustomHeaderButton(string.Empty, Properties.Resources.Undo);

                dockPanel.CustomHeaderButtons.Add(button);

                if(contentControl != null)
                {
                    dockPanel.ControlContainer.Controls.Add(contentControl);
                }

                if(dockStyle == DockStyle.Fill || dockStyle == DockStyle.None)
                {
                    try
                    {
                        dockPanel.DockedAsTabbedDocument = true;

                        Document document = this.tabbedView.AddDocument(dockPanel) as Document;

                        //RefreshScreen();

                        if(documentOption != null)
                        {
                            if(documentOption.FloatLocation.HasValue)
                            {
                                document.FloatLocation = documentOption.FloatLocation;
                            }

                            if(documentOption.FloatSize.HasValue)
                            {
                                document.FloatSize = documentOption.FloatSize;
                            }
                        }
                    }
                    finally
                    {
                        this.tabbedView.EndUpdate();
                    }
                }

                dockPanel.CustomButtonClick += dockPanel_CustomButtonClick;
                dockPanel.ClosedPanel       += dockPanel_ClosedPanel;

                this.dictionary.Add(name, dockPanel);

                return true;
            }
            finally
            {
                this.isAddDocument = false;

                this.dockManager.EndUpdate();
                this.tabbedView.EndUpdate();
            }
        }

        #endregion
        #region 문서 수 구하기 - GetDocumentCount()

        /// <summary>
        /// 문서 수 구하기
        /// </summary>
        /// <returns>문서 수</returns>
        public int GetDocumentCount()
        {
            int count = this.dockManager.Panels.Count;

            return count;
        }

        #endregion
        #region 문서 옵션 구하기 - GetDocumentOption(name)

        /// <summary>
        /// 문서 옵션 구하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>문서 옵션</returns>
        public DocumentOption GetDocumentOption(string name)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                DocumentOption documentOption = new DocumentOption();

                SetDocumentOption(documentOption, dockPanel);

                return documentOption;
            }
            else
            {
                return null;
            }
        }

        #endregion
        #region 문서 옵션 구하기 - GetDocumentOption(index)

        /// <summary>
        /// 문서 옵션 구하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>문서 옵션</returns>
        public DocumentOption GetDocumentOption(int index)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            DocumentOption documentOption = GetDocumentOption(dockPanel.Name);

            return documentOption;
        }

        #endregion
        #region 문서 옵션 설정하기 - ion(name, documentOption)

        /// <summary>
        /// 문서 옵션 설정하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <param name="documentOption">문서 옵션</param>
        /// <returns>처리 결과</returns>
        public bool SetDocumentOption(string name, DocumentOption documentOption)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                SetDockPanel(dockPanel, documentOption);

                return true;
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 문서 옵션 설정하기 - SetDocumentOption(index, documentOption)

        /// <summary>
        /// 문서 옵션 설정하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <param name="documentOption">문서 옵션</param>
        /// <returns>처리 결과</returns>
        public bool SetDocumentOption(int index, DocumentOption documentOption)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = SetDocumentOption(dockPanel.Name, documentOption);

            return result;
        }

        #endregion
        #region 컨텐트 컨트롤 구하기 - GetContentControl(name)

        /// <summary>
        /// 컨텐트 컨트롤 구하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>컨텐트 컨트롤</returns>
        public Control GetContentControl(string name)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                if(dockPanel.ControlContainer.Controls.Count > 0)
                {
                    return dockPanel.ControlContainer.Controls[0];
                }
                else
                {
                    return null;
                }
            }
            else
            {
                return null;
            }
        }

        #endregion
        #region 컨텐트 컨트롤 구하기 - GetContentControl(index)

        /// <summary>
        /// 컨텐트 컨트롤 구하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>컨텐트 컨트롤</returns>
        public Control GetContentControl(int index)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            Control control = GetContentControl(dockPanel.Name);

            return control;
        }

        #endregion
        #region 문서 제거하기 - RemoveDocument(name)

        /// <summary>
        /// 문서 제거하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>처리 결과</returns>
        public bool RemoveDocument(string name)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                dockPanel.Close();

                return true;
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 문서 제거하기 - RemoveDocument(index)

        /// <summary>
        /// 문서 제거하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>처리 결과</returns>
        public bool RemoveDocument(int index)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = RemoveDocument(dockPanel.Name);

            return result;
        }

        #endregion
        #region 모든 문서들 제거하기 - RemoveAllDocuments()

        /// <summary>
        /// 모든 문서들 제거하기
        /// </summary>
        public void RemoveAllDocuments()
        {
            try
            {
                this.dockManager.BeginUpdate();

                this.dockManager.Clear();

                this.dictionary.Clear();
            }
            finally
            {
                this.dockManager.EndUpdate();
            }
        }

        #endregion
        #region 플로팅 문서 여부 구하기 - IsFloatDocument(name)

        /// <summary>
        /// 플로팅 문서 여부 구하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>플로팅 문서 여부</returns>
        public bool IsFloatDocument(string name)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                return dockPanel.FloatForm != null;
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 플로팅 문서 여부 구하기 - IsFloatDocument(index)

        /// <summary>
        /// 플로팅 문서 여부 구하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>플로팅 문서 여부</returns>
        public bool IsFloatDocument(int index)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = IsFloatDocument(dockPanel.Name);

            return result;
        }

        #endregion
        #region 플로팅 문서 만들기 - ToFloatDocument(name, floatLocation)

        /// <summary>
        /// 플로팅 문서 만들기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <param name="floatLocation">플로팅 위치</param>
        /// <returns>처리 결과</returns>
        public bool ToFloatDocument(string name, Point? floatLocation = null)
        {
            if(this.dictionary.ContainsKey(name))
            {
                if(IsFloatDocument(name))
                {
                    return false;
                }

                DockPanel dockPanel = this.dictionary[name];

                BaseDocument document = GetDocument(dockPanel);

                if(document == null)
                {
                    dockPanel.MakeFloat();

                    return true;
                }
                else
                {
                    this.documentManager.View.Controller.Float(document);

                    return true;
                }
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 플로팅 문서 만들기 - ToFloatDocument(index, floatLocation)

        /// <summary>
        /// 플로팅 문서 만들기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <param name="floatLocation">플로팅 위치</param>
        /// <returns>처리 결과</returns>
        public bool ToFloatDocument(int index, Point? floatLocation = null)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = ToFloatDocument(dockPanel.Name, floatLocation);

            return result;
        }

        #endregion
        #region 탭 문서 만들기 - ToTabbedDocument(name)

        /// <summary>
        /// 탭 문서 만들기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <returns>처리 결과</returns>
        public bool ToTabbedDocument(string name)
        {
            if(this.dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                if(IsFloatDocument(dockPanel.Name))
                {
                    try
                    {
                        this.tabbedView.BeginUpdate();

                        dockPanel.DockedAsTabbedDocument = true;

                        this.tabbedView.AddDocument(dockPanel);
                    }
                    finally
                    {
                        this.tabbedView.EndUpdate();
                    }

                    return true;
                }
                else
                {
                    BaseDocument document = GetDocument(dockPanel);

                    if(document == null)
                    {
                        try
                        {
                            this.tabbedView.BeginUpdate();

                            dockPanel.DockedAsTabbedDocument = true;

                            this.tabbedView.AddDocument(dockPanel);
                        }
                        finally
                        {
                            this.tabbedView.EndUpdate();
                        }

                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 탭 문서 만들기 - ToTabbedDocument(index)

        /// <summary>
        /// 탭 문서 만들기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <returns>처리 결과</returns>
        public bool ToTabbedDocument(int index)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = ToTabbedDocument(dockPanel.Name);

            return result;
        }

        #endregion
        #region 탭 문서 이동하기 - MoveTabbedDocument(name, position)

        /// <summary>
        /// 탭 문서 이동하기
        /// </summary>
        /// <param name="name">명칭</param>
        /// <param name="position">위치</param>
        /// <returns>처리 결과</returns>
        public bool MoveTabbedDocument(string name, int position)
        {
            if(dictionary.ContainsKey(name))
            {
                DockPanel dockPanel = this.dictionary[name];

                if(IsFloatDocument(dockPanel.Name))
                {
                    return false;
                }
                else
                {
                    Document document = GetDocument(dockPanel) as Document;

                    if(document == null)
                    {
                        return false;
                    }
                    else
                    {
                        this.tabbedView.Controller.Move(document, position);

                        return true;
                    }
                }
            }
            else
            {
                return false;
            }
        }

        #endregion
        #region 탭 문서 이동하기 - MoveTabbedDocument(index, position)

        /// <summary>
        /// 탭 문서 이동하기
        /// </summary>
        /// <param name="index">인덱스</param>
        /// <param name="position">위치</param>
        /// <returns>처리 결과</returns>
        public bool MoveTabbedDocument(int index, int position)
        {
            DockPanel dockPanel = this.dockManager.Panels[index];

            bool result = MoveTabbedDocument(dockPanel.Name, position);

            return result;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 탭 뷰 문서 추가시 처리하기 - tabbedView_DocumentAdded(sender, e)

        /// <summary>
        /// 탭 뷰 문서 추가시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void tabbedView_DocumentAdded(object sender, DocumentEventArgs e)
        {
            if(!this.isAddDocument)
            {
                RefreshScreen();
            }
        }

        #endregion
        #region 도킹 패널 커스텀 버튼 클릭시 처리하기 - dockPanel_CustomButtonClick(sender, e)

        /// <summary>
        /// 도킹 패널 커스텀 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void dockPanel_CustomButtonClick(object sender, ButtonEventArgs e)
        {
            DockPanel dockPanel = sender as DockPanel;

            try
            {
                this.tabbedView.BeginUpdate();

                dockPanel.DockedAsTabbedDocument = true;

                Document document = this.tabbedView.AddDocument(dockPanel) as Document;

                RefreshScreen();
            }
            finally
            {
                this.tabbedView.EndUpdate();
            }
        }

        #endregion
        #region 도킹 패널 패널 닫힌 경우 처리하기 - dockPanel_ClosedPanel(sender, e)

        /// <summary>
        /// 도킹 패널 패널 닫힌 경우 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void dockPanel_ClosedPanel(object sender, DockPanelEventArgs e)
        {
            DockPanel dockPanel = sender as DockPanel;

            if(this.dictionary.ContainsKey(dockPanel.Name))
            {
                this.dictionary.Remove(dockPanel.Name);
            }

            this.dockManager.RemovePanel(dockPanel);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 커스텀 헤더 버튼 구하기 - GetCustomHeaderButton(caption, image)

        /// <summary>
        /// 커스텀 헤더 버튼 구하기
        /// </summary>
        /// <param name="caption">제목</param>
        /// <param name="image">이미지</param>
        /// <returns>커스텀 헤더 버튼</returns>
        private CustomHeaderButton GetCustomHeaderButton(string caption, Image image)
        {
            CustomHeaderButton button = new CustomHeaderButton(caption, DevExpress.XtraBars.Docking2010.ButtonStyle.PushButton);

            button.Image = image;

            return button;
        }

        #endregion
        #region 도킹 패널 설정하기 - SetDockPanel(dockPanel, documentOption)

        /// <summary>
        /// 도킹 패널 설정하기
        /// </summary>
        /// <param name="dockPanel">도킹 패널</param>
        /// <param name="documentOption">문서 옵션</param>
        /// <returns>처리 결과</returns>
        private bool SetDockPanel(DockPanel dockPanel, DocumentOption documentOption)
        {
            if(dockPanel == null || documentOption == null)
            {
                return false;
            }

            dockPanel.Options.AllowDockLeft      = documentOption.AllowDockLeft;
            dockPanel.Options.AllowDockTop       = documentOption.AllowDockTop;
            dockPanel.Options.AllowDockRight     = documentOption.AllowDockRight;
            dockPanel.Options.AllowDockBottom    = documentOption.AllowDockBottom;
            dockPanel.Options.AllowDockFill      = documentOption.AllowDockFill;
            dockPanel.Options.AllowFloating      = documentOption.AllowFloating;
            dockPanel.Options.FloatOnDblClick    = documentOption.FloatOnDblClick;
            dockPanel.Options.ShowAutoHideButton = documentOption.ShowAutoHideButton;
            dockPanel.Options.ShowMaximizeButton = documentOption.ShowMaximizeButton;
            dockPanel.Options.ShowCloseButton    = documentOption.ShowCloseButton;

            BaseDocument document = GetDocument(dockPanel);

            if(document != null)
            {
                if(documentOption.FloatLocation.HasValue)
                {
                    document.FloatLocation = documentOption.FloatLocation;
                }

                if(documentOption.FloatSize.HasValue)
                {
                    document.FloatSize = documentOption.FloatSize;
                }
            }

            return true;
        }

        #endregion
        #region 문서 옵션 설정하기 - SetDocumentOption(documentOption, dockPanel)

        /// <summary>
        /// 문서 옵션 설정하기
        /// </summary>
        /// <param name="documentOption">문서 옵션</param>
        /// <param name="dockPanel">도킹 패널</param>
        /// <returns>처리 결과</returns>
        private bool SetDocumentOption(DocumentOption documentOption, DockPanel dockPanel)
        {
            if(documentOption == null || dockPanel == null)
            {
                return false;
            }

            documentOption.AllowDockLeft      = dockPanel.Options.AllowDockLeft;
            documentOption.AllowDockTop       = dockPanel.Options.AllowDockTop;
            documentOption.AllowDockRight     = dockPanel.Options.AllowDockRight;
            documentOption.AllowDockBottom    = dockPanel.Options.AllowDockBottom;
            documentOption.AllowDockFill      = dockPanel.Options.AllowDockFill;
            documentOption.AllowFloating      = dockPanel.Options.AllowFloating;
            documentOption.FloatOnDblClick    = dockPanel.Options.FloatOnDblClick;
            documentOption.ShowAutoHideButton = dockPanel.Options.ShowAutoHideButton;
            documentOption.ShowMaximizeButton = dockPanel.Options.ShowMaximizeButton;
            documentOption.ShowCloseButton    = dockPanel.Options.ShowCloseButton;

            BaseDocument document = GetDocument(dockPanel);

            if(document != null)
            {
                documentOption.FloatLocation = document.FloatLocation;
                documentOption.FloatSize     = document.FloatSize;
            }
            else
            {
                documentOption.FloatLocation = null;
                documentOption.FloatSize     = null;
            }

            return true;
        }

        #endregion
        #region 문서 구하기 - GetDocument(control)

        /// <summary>
        /// 문서 구하기
        /// </summary>
        /// <param name="control">컨트롤</param>
        /// <returns>문서</returns>
        private BaseDocument GetDocument(Control control)
        {
            BaseDocument document = this.documentManager.GetDocument(control);

            return document;
        }

        #endregion
        #region 도킹 스타일 구하기 - GetDockingStyle(dockStyle)

        /// <summary>
        /// 도킹 스타일 구하기
        /// </summary>
        /// <param name="dockStyle">도킹 스타일</param>
        /// <returns>도킹 스타일</returns>
        private DockingStyle GetDockingStyle(DockStyle dockStyle)
        {
            switch(dockStyle)
            {
                case DockStyle.Left   : return DockingStyle.Left;
                case DockStyle.Top    : return DockingStyle.Top;
                case DockStyle.Right  : return DockingStyle.Right;
                case DockStyle.Bottom : return DockingStyle.Bottom;
                default               : return DockingStyle.Float;
            }
        }

        #endregion
        #region 화면 갱신하기 - RefreshScreen()

        /// <summary>
        /// 화면 갱신하기
        /// </summary>
        private void RefreshScreen()
        {
            InvalidateRect(IntPtr.Zero, IntPtr.Zero, true);
        }

        #endregion
    }
}